import java.util.ArrayList;
import java.util.Iterator;
import java.util.Set;
import java.util.TreeMap;
import java.util.Map.Entry;
import java.util.concurrent.atomic.AtomicInteger;

public class OTBTree implements IntSet {

	final int ADD = 0;
	final int REMOVE = 1;
	final int CONTAINS = 2;

	int lockAbort = 0;
	int validateAbort = 0;

	int num_removes = 0;
	int num_rotations = 0;

	public static final int NOT_REMOVED = 0;
	public static final int BY_LEFT_ROT = 1;
	public static final int REMOVED = 2;

	static final boolean useFairLocks = false;
	// we encode directions as characters
	static final char Left = 'L';
	static final char Right = 'R';

	private class MaintenanceThread extends Thread {
		OTBTree set;

		MaintenanceThread(OTBTree set) {
			this.set = set;
		}

		public void run() {
			set.doMaintenance();
		}
	}

	private class MaintVariables {
		long propogations = 0, rotations = 0;
	}

	private final MaintVariables vars = new MaintVariables();

	private static class Node {
		int key;

		volatile AtomicInteger lock = new AtomicInteger(0);
		volatile long lockHolder = -1;

		class BalanceVars {
			volatile int localh, lefth, righth;
		}

		final BalanceVars bal = new BalanceVars();
		volatile Node left;
		volatile Node right;
		// final ReentrantLock lock;
		volatile int removed;
		volatile boolean deleted;

		Node(final int key) {
			this.key = key;
			this.removed = NOT_REMOVED;
			this.deleted = false;
			// this.lock = new ReentrantLock(useFairLocks);
			this.right = null;
			this.left = null;
			this.bal.localh = 1;
			this.bal.righth = 0;
			this.bal.lefth = 0;
		}

		Node(final int key, final int localh, final int lefth,
				final int righth, final Node left, final Node right,
				int removed, boolean deleted) {
			this.key = key;
			this.bal.localh = localh;
			this.bal.righth = righth;
			this.bal.lefth = lefth;
			this.left = left;
			this.right = right;
			// this.lock = new ReentrantLock(useFairLocks);
			this.removed = removed;
			this.deleted = deleted;
		}

		void updateLocalh() {
			this.bal.localh = Math.max(this.bal.lefth + 1, this.bal.righth + 1);
		}

	}

	// state
	private final Node root = new Node(Integer.MIN_VALUE);
	volatile boolean stop = false;
	private MaintenanceThread mainThd;
	// used in the getSize function
	int size;

	// Constructors
	public OTBTree() {
		// temporary
		this.startMaintenance();
	}

	Node getNext(Node n, int k) {
		int rem = n.removed;
		if (rem == BY_LEFT_ROT)
			return n.right;
		else if (rem == REMOVED)
			return n.left;
		else if (n.key > k)
			return n.right;
		else if (n.key < k)
			return n.left;
		else
			// n.key == k
			return null;
	}

	boolean validate(Node n, int k, boolean itemExists) {
		Node next = null;
		if (n.removed != NOT_REMOVED)
			return false;
		else if (n.key == k) {
			if (n.deleted != itemExists)
				return true;
			else
				return false;
		} else if (n.key > k)
			next = n.right;
		else
			next = n.left;

		if (next == null && !itemExists)
			return true;
		return false;
	}

	class ReadSetEntry {
		Node node;
		int k;
		boolean itemExists;

		public ReadSetEntry(Node node, int k, boolean itemExists) {
			this.node = node;
			this.k = k;
			this.itemExists = itemExists;
		}
	}

	class WriteSetEntry {
		Node node;
		int k;
		Node newNode;
		int operation;

		public WriteSetEntry(Node node, int k, int operation) {
			this.node = node;
			this.k = k;
			this.operation = operation;
		}
	}

	@Override
	public boolean add(int item) throws AbortedException {
		return operation(ADD, item);
	}

	@Override
	public boolean remove(int item) throws AbortedException {
		return operation(REMOVE, item);
	}

	@Override
	public boolean contains(int item) throws AbortedException {
		return operation(CONTAINS, item);
	}

	private boolean commitValidate(ArrayList<ReadSetEntry> readset, long threadId) {
//		ReadSetEntry entry;
//		int size = readset.size();
//
//		// check the values of the nodes
//		for (int i = 0; i < size; i++) {
//			entry = readset.get(i);
//			if (!validate(entry.node, entry.k, entry.itemExists))
//				return false;
//		}
//		return true;
		return postValidate(readset, threadId);
	}

	private boolean postValidate(ArrayList<ReadSetEntry> readset, long threadId) {
		
		ReadSetEntry entry;
		int size = readset.size();

		// check the values of the nodes
		// and also check that nodes are not currently locked
		for (int i = 0; i < size; i++) {
			entry = readset.get(i);
			if((entry.node.lock.get() & 1) == 1 && (entry.node.lockHolder != threadId))
				return false;
			if (!validate(entry.node, entry.k, entry.itemExists))
				return false;
		}
		return true;
	}

	private boolean operation(int type, int item) throws AbortedException {
		SetThread t = ((SetThread) Thread.currentThread());
		TreeMap<Integer, WriteSetEntry> writeset = t.tree_writeset;
		long threadId = Thread.currentThread().getId();

		WriteSetEntry entry = writeset.get(item);
		if (entry != null) {
			if (entry.operation == ADD) {
				if (type == ADD)
					return false;
				else if (type == CONTAINS)
					return true;
				else // remove
				{
					// System.out.println("remove after add");
					writeset.remove(item);
					return true;
				}
			} else // remove
			{
				if (type == REMOVE || type == CONTAINS)
					return false;
				else // add
				{
					// System.out.println("add after remove");
					writeset.remove(item);
					return true;
				}
			}
		}

		ArrayList<ReadSetEntry> readset = ((SetThread) Thread.currentThread()).tree_readset;

		Node next, node;
		node = root;
		int c = 0;
		while (true) {
			next = getNext(node, item);
			if (next == null)
				break;
			node = next;
//			c++;
//			if(c > 20)
//				System.out.println("Too long tree\n");
		}

		int currItem = node.key;
		boolean currDeleted = node.deleted;

		boolean returnValue = false;

		if (currItem == item && !currDeleted) {
			readset.add(new ReadSetEntry(node, item, true));
			if (type == CONTAINS) {
				returnValue = true;
			} else if (type == ADD) {
				returnValue = false;
			} else // remove
			{
				writeset.put(item, new WriteSetEntry(node, item, REMOVE));
				returnValue = true;
			}
		} else {
			readset.add(new ReadSetEntry(node, item, false));
			if (type == CONTAINS || type == REMOVE)
				returnValue = false;
			else // add
			{
				writeset.put(item, new WriteSetEntry(node, item, ADD));
				returnValue = true;
			}
		}

		if (!postValidate(readset, threadId))
			throw AbortedException.abortedException;

		return returnValue;
	}

	@Override
	public void commit() throws AbortedException {
		SetThread t = ((SetThread) Thread.currentThread());

		Set<Entry<Integer, WriteSetEntry>> write_set = t.tree_writeset
				.entrySet();
		ArrayList<ReadSetEntry> read_set = t.tree_readset;

		// read-only transactions do nothing
		if (write_set.isEmpty()) {
			read_set.clear();
			return;
		}

		long threadId = Thread.currentThread().getId();
		Iterator<Entry<Integer, WriteSetEntry>> iterator = write_set.iterator();
		WriteSetEntry entry;

		int lock;
		Node node, next, newNode;

		// Acquire locks
		while (iterator.hasNext()) {
			entry = iterator.next().getValue();

			lock = entry.node.lock.get();

			// check that pred lock is not acquired by another thread
			if ((lock & 1) == 1 && entry.node.lockHolder != threadId)
				throw AbortedException.abortedException;

//			while((lock & 1) == 1 && entry.node.lockHolder != threadId)
//				lock = entry.node.lock.get();
			
			// Try to acquire pred lock
			if (entry.node.lockHolder == threadId
					|| entry.node.lock.compareAndSet(lock, lock + 1)) {
				entry.node.lockHolder = threadId;
			} else
				throw AbortedException.abortedException;
		}

		// validate read-set
		if (!commitValidate(t.tree_readset, threadId))
			throw AbortedException.abortedException;

		// Publish write-set
		iterator = write_set.iterator();
		while (iterator.hasNext()) {
			entry = iterator.next().getValue();

			if (entry.operation == ADD) {
				if (entry.node.key == entry.k) {
					if (!entry.node.deleted
							|| entry.node.removed != NOT_REMOVED)
						System.out.println("CANNOT HAPPEN\n");
					entry.node.deleted = false;
				} else {
					newNode = new Node(entry.k);
					newNode.lock.set(1);
					newNode.lockHolder = threadId;
					entry.newNode = newNode;

					// keep iterating. The only case here is that other
					// nodes
					// are inserted in the same transaction.
					node = entry.node;
					while (true) {
						next = getNext(node, entry.k);
						if (next == null) {
							break;
						} else
							node = next;
					}

					if (node.key > entry.k)
						node.right = newNode;
					else
						node.left = newNode;
				}
			} else // remove
			{
				if(entry.node.key != entry.k)
					System.out.println("CANNOT HAPPEN\n");
				entry.node.deleted = true;
			}
		}

		// unlock
		iterator = write_set.iterator();
		while (iterator.hasNext()) {
			entry = iterator.next().getValue();
			if (entry.node.lockHolder == threadId) {
				entry.node.lockHolder = -1;
				entry.node.lock.decrementAndGet();
			}
			// Unlock newNode
			if (entry.operation == ADD && entry.newNode != null) {
				if (entry.newNode.lockHolder == threadId) {
					entry.newNode.lockHolder = -1;
					entry.newNode.lock.decrementAndGet();
				}
			}
		}

		// clear read- and write- sets
		read_set.clear();
		write_set.clear();
	}

	@Override
	public void abort() {

		SetThread t = ((SetThread) Thread.currentThread());

		// internally you must not call this method. It's sufficient to throw
		// AbortException
		// This method will be called from calling transaction as a part of its
		// abort subroutine
		// This way: Abort is handled by transaction not by data structure.

		// Unlock
		Iterator<Entry<Integer, WriteSetEntry>> iterator = t.tree_writeset
				.entrySet().iterator();
		WriteSetEntry entry;
		while (iterator.hasNext()) {
			entry = iterator.next().getValue();
			if (entry.node.lockHolder == t.getId()) {
				entry.node.lockHolder = -1;
				entry.node.lock.decrementAndGet();
			}
		}

		// clear read- and write- sets
		t.tree_readset.clear();
		t.tree_writeset.clear();

	}

	// maintenance

	void maintenanceLock(Node node) {
		long threadId = Thread.currentThread().getId();
		int lock;
		while (true) {
			lock = node.lock.get();
			if ((lock & 1) == 1)
				continue;
			if (node.lock.compareAndSet(lock, lock + 1)) {
				node.lockHolder = threadId;
				break;
			}
		}
	}

	void maintenanceUnlock(Node node, boolean rollBack) {
		node.lockHolder = -1;
		if(rollBack)
			node.lock.getAndDecrement();
		else
			node.lock.getAndDecrement();
	}

	boolean removeNode(Node parent, char direction) {
		Node n, child;
		// can get before locks because only maintenance removes nodes
		if (parent.removed != NOT_REMOVED)
			return false;
		n = direction == Left ? parent.left : parent.right;
		if (n == null)
			return false;
		// get the locks
		maintenanceLock(n);
		maintenanceLock(parent);
		if (!n.deleted) {
			maintenanceUnlock(n, true);
			maintenanceUnlock(parent, true);
			return false;
		}
		if ((child = n.left) != null) {
			if (n.right != null) {
				maintenanceUnlock(n, true);
				maintenanceUnlock(parent, true);
				return false;
			}
		} else {
			child = n.right;
		}
		if (direction == Left) {
			parent.left = child;
		} else {
			parent.right = child;
		}
		n.left = parent;
		n.right = parent;

		num_removes++;

		n.removed = REMOVED;
		maintenanceUnlock(n, false);
		maintenanceUnlock(parent, false);
		// System.out.println("removed a node");
		// need to update balance values here
		if (direction == Left) {
			parent.bal.lefth = n.bal.localh - 1;
		} else {
			parent.bal.righth = n.bal.localh - 1;
		}
		parent.updateLocalh();
		return true;
	}

	int rightRotate(Node parent, char direction, boolean doRotate) {

		Node n, l, lr, r, newNode;
		if (parent.removed != NOT_REMOVED)
			return 0;
		n = direction == Left ? parent.left : parent.right;
		if (n == null)
			return 0;
		l = n.left;
		if (l == null)
			return 0;
		if (l.bal.lefth - l.bal.righth < 0 && !doRotate) {
			// should do a double rotate
			return 2;
		}
		maintenanceLock(parent);
		maintenanceLock(n);
		maintenanceLock(l);
		lr = l.right;
		r = n.right;
		newNode = new Node(n.key, Math.max(1 + l.bal.righth, 1 + n.bal.righth),
				l.bal.righth, n.bal.righth, lr, r, n.removed, n.deleted);
		l.right = newNode;

		// num_rotations++;

		n.removed = REMOVED;
		if (direction == Left) {
			parent.left = l;
		} else {
			parent.right = l;
		}
		maintenanceUnlock(l, false);
		maintenanceUnlock(n, false);
		maintenanceUnlock(parent, false);
		// need to update balance values
		l.bal.righth = newNode.bal.localh;
		l.updateLocalh();
		if (direction == Left) {
			parent.bal.lefth = l.bal.localh;
		} else {
			parent.bal.righth = l.bal.localh;
		}
		parent.updateLocalh();
		// System.out.println("right rotate");
		return 1;

	}

	int leftRotate(Node parent, char direction, boolean doRotate) {
		Node n, r, rl, l, newNode;
		if (parent.removed != NOT_REMOVED)
			return 0;
		n = direction == Left ? parent.left : parent.right;
		if (n == null)
			return 0;
		r = n.right;
		if (r == null)
			return 0;
		if (r.bal.lefth - r.bal.righth > 0 && !doRotate) {
			// should do a double rotate
			return 3;
		}
		maintenanceLock(parent);
		maintenanceLock(n);
		maintenanceLock(r);
		rl = r.left;
		l = n.left;

		newNode = new Node(n.key, Math.max(1 + r.bal.lefth, 1 + n.bal.lefth),
				n.bal.lefth, r.bal.lefth, l, rl, n.removed, n.deleted);

		r.left = newNode;

		// // temp (Need to fix this!!!!!!!!!!!!!!!!!!!!)
		// n.right = parent;
		// n.left = parent;

		n.removed = BY_LEFT_ROT;

		num_rotations++;

		if (direction == Left) {
			parent.left = r;
		} else {
			parent.right = r;
		}
		maintenanceUnlock(r, false);
		maintenanceUnlock(n, false);
		maintenanceUnlock(parent, false);
		// need to update balance values
		r.bal.righth = newNode.bal.localh;
		r.updateLocalh();
		if (direction == Left) {
			parent.bal.lefth = r.bal.localh;
		} else {
			parent.bal.righth = r.bal.localh;
		}
		parent.updateLocalh();
		// System.out.println("left rotate");
		return 1;
	}

	boolean propagate(Node node) {
		Node lchild, rchild;

		lchild = node.left;
		rchild = node.right;

		if (lchild == null) {
			node.bal.lefth = 0;
		} else {
			node.bal.lefth = lchild.bal.localh;
		}
		if (rchild == null) {
			node.bal.righth = 0;
		} else {
			node.bal.righth = rchild.bal.localh;
		}

		node.updateLocalh();

		if (Math.abs(node.bal.righth - node.bal.lefth) >= 2)
			return true;
		return false;
	}

	boolean performRotation(Node parent, char direction) {
		int ret;
		Node node;

		ret = singleRotation(parent, direction, false, false);
		if (ret == 2) {
			// Do a LRR
			node = direction == Left ? parent.left : parent.right;
			ret = singleRotation(node, Left, true, false);
			if (ret > 0) {
				if (singleRotation(parent, direction, false, true) > 0) {
					// System.out.println("LRR");
				}
			}
		} else if (ret == 3) {
			// Do a RLR
			node = direction == Left ? parent.left : parent.right;
			ret = singleRotation(node, Right, false, true);
			if (ret > 0) {
				if (singleRotation(parent, direction, true, false) > 0) {
					// System.out.println("RLR");
				}
			}
		}
		if (ret > 0)
			return true;
		return false;
	}

	int singleRotation(Node parent, char direction, boolean leftRotation,
			boolean rightRotation) {
		int bal, ret = 0;
		Node node, child;

		node = direction == Left ? parent.left : parent.right;
		bal = node.bal.lefth - node.bal.righth;
		if (bal >= 2 || rightRotation) {
			// check reiable and rotate
			child = node.left;
			if (child != null) {
				if (node.bal.lefth == child.bal.localh) {
					ret = rightRotate(parent, direction, rightRotation);
				}
			}
		} else if (bal <= -2 || leftRotation) {
			// check reliable and rotate
			child = node.right;
			if (child != null) {
				if (node.bal.righth == child.bal.localh) {
					ret = leftRotate(parent, direction, leftRotation);
				}
			}
		}
		return ret;
	}

	boolean recursivePropagate(Node parent, Node node, char direction) {

		Node left, right;

		if (node == null)
			return true;
		left = node.left;
		right = node.right;

		if (node.removed == NOT_REMOVED && node.deleted
				&& (left == null || right == null) && node != this.root) {
			if (removeNode(parent, direction)) {
				return true;
			}
		}

		if (stop) {
			return true;
		}

		if (node.removed == NOT_REMOVED) {
			if (left != null) {
				recursivePropagate(node, left, Left);
			}
			if (right != null) {
				recursivePropagate(node, right, Right);
			}
		}

		if (stop) {
			return true;
		}

		// no rotations for now
		if (node.removed == NOT_REMOVED && node != this.root) {
			if (propagate(node)) {
				this.performRotation(parent, direction);
			}
		}

		return true;
	}

	public boolean stopMaintenance() {
		this.stop = true;
		try {
			this.mainThd.join();
		} catch (InterruptedException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		System.out.println("\nNumber of rotations: " + num_rotations + "\nNumber of physical removals: " + num_removes + "\n");
		return true;
	}

	public boolean startMaintenance() {
		this.stop = false;

		mainThd = new MaintenanceThread(this);

		mainThd.start();

		return true;
	}

	boolean doMaintenance() {
		while (!stop) {
			recursivePropagate(this.root, this.root.left, Left);
		}
		return true;
	}

	// not thread safe
	public int getSize() {
		this.size = 0;
		recursiveGetSize(root.left);
		return size;
	}

	void recursiveGetSize(Node node) {
		if (node == null)
			return;
		if (node.removed != NOT_REMOVED) {
			// System.out.println("Shouldn't find removed nodes in the get size function");
		}
		if (!node.deleted) {
			this.size++;
		}
		recursiveGetSize(node.left);
		recursiveGetSize(node.right);
	}

	public int getBalance() {
		int lefth = 0, righth = 0;
		if (root.left == null)
			return 0;
		lefth = recursiveDepth(root.left.left);
		righth = recursiveDepth(root.left.right);
		return lefth - righth;
	}

	int recursiveDepth(Node node) {
		if (node == null) {
			return 0;
		}
		int lefth, righth;
		lefth = recursiveDepth(node.left);
		righth = recursiveDepth(node.right);
		return Math.max(lefth, righth) + 1;
	}

	public void clear() {
		this.stopMaintenance();
		this.resetTree();
		this.startMaintenance();

		return;
	}

	private void resetTree() {
		this.vars.propogations = 0;
		this.vars.rotations = 0;
		root.left = null;
	}

	@Override
	public void begin() {
		// TODO Auto-generated method stub

	}

	// This is a trick to insert the items balanced from the beginning
	// It's important to save the unnecessary time needed to re-balance
	// which is significant if the range is large.
	public boolean nontransactionalAdd(int key) {
		Node next, node;
		node = root;

		while (true) {
			next = getNext(node, key);
			if (next == null) {
				maintenanceLock(node);
				if (validate(node, key, false))
					break;
				maintenanceUnlock(node, false);
			} else
				node = next;
		}

		boolean result = false;

		if (node.key == key) {
			if (node.deleted) {
				node.deleted = false;
				result = true;
			}
		} else {
			Node n = new Node(key);
			if (node.key > key)
				node.right = n;
			else
				node.left = n;
			result = true;
		}
		maintenanceUnlock(node, false);
		return result;
	}

	public void recprint(Node p) {
		if (p == null)
			return;
		recprint(p.right);
		System.out.print(p.key + "  ");
		recprint(p.left);
	}

	public int recgetvalue(Node p) {
		if (p == null)
			return 0;
		if (p.deleted || p == root)
			return recgetvalue(p.right) + recgetvalue(p.left);
		else
			return recgetvalue(p.right) + p.key + recgetvalue(p.left);
	}

	public void recursiveAdd(int start, int end) {
		if (start > end)
			return;

		int mid = (start + end) / 2;
		nontransactionalAdd(mid * 2);
		recursiveAdd(start, mid - 1);
		recursiveAdd(mid + 1, end);

		return;
	}

	@Override
	public int finish() {
		// TODO Auto-generated method stub
		stopMaintenance();
		return recgetvalue(root);
	}

	@Override
	public void start(int range) {
		// TODO Auto-generated method stub
		recursiveAdd(0, range / 2 - 1);

		// recprint(root);
		// System.out.println();
	}

}
