import java.util.ArrayList;
import java.util.Iterator;
import java.util.Set;
import java.util.TreeMap;
import java.util.Map.Entry;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicInteger;

public class OTBTreeOptimizedBackoff implements IntSet {

	final int ADD = 6;
	final int REMOVE = 7;
	final int CONTAINS = 8;

	int lockAbort = 0;
	int validateAbort = 0;

	int num_removes = 0;
	int num_rotations = 0;

	public static final int MAINTAINANCE_THREAD_ID = 1000;

	public static final int NOT_REMOVED = 0;
	public static final int BY_LEFT_ROT = 1;
	public static final int REMOVED = 2;

	public static final int STRUCTURALLY_INVALID = 0;
	public static final int SEMANTICALLY_INVALID = 1;
	public static final int VALID = 2;

	public static final int UNLOCKED = 0;
	public static final int LOCKED = 1;

	public static final int STRUCTURAL_LOCK = 0;
	public static final int SEMANTIC_LOCK = 1;
	public static final int BOTH_LOCKS = 2;

	static final boolean useFairLocks = false;
	// we encode directions as characters
	static final char Left = 'L';
	static final char Right = 'R';
	
	volatile int delayCount = 0;

	private class MaintenanceThread extends Thread {
		OTBTreeOptimizedBackoff set;

		MaintenanceThread(OTBTreeOptimizedBackoff set) {
			this.set = set;
		}

		public void run() {
			set.doMaintenance();
		}
	}

	private class MaintVariables {
		long propogations = 0, rotations = 0;
	}

	private final MaintVariables vars = new MaintVariables();

	private static class Node {
		int key;

		volatile AtomicInteger semantic_lock = new AtomicInteger(0);
		volatile long semantic_lockHolder = -1;
		volatile AtomicInteger struct_lock = new AtomicInteger(0);
		volatile long struct_lockHolder = -1;

		class BalanceVars {
			volatile int localh, lefth, righth;
		}

		final BalanceVars bal = new BalanceVars();
		volatile Node left;
		volatile Node right;
		// final ReentrantLock lock;
		volatile int removed;
		volatile boolean deleted;

		Node(final int key) {
			this.key = key;
			this.removed = NOT_REMOVED;
			this.deleted = false;
			// this.lock = new ReentrantLock(useFairLocks);
			this.right = null;
			this.left = null;
			this.bal.localh = 1;
			this.bal.righth = 0;
			this.bal.lefth = 0;
		}

		Node(final int key, final int localh, final int lefth,
				final int righth, final Node left, final Node right,
				int removed, boolean deleted) {
			this.key = key;
			this.bal.localh = localh;
			this.bal.righth = righth;
			this.bal.lefth = lefth;
			this.left = left;
			this.right = right;
			// this.lock = new ReentrantLock(useFairLocks);
			this.removed = removed;
			this.deleted = deleted;
		}

		void updateLocalh() {
			this.bal.localh = Math.max(this.bal.lefth + 1, this.bal.righth + 1);
		}

	}

	// state
	private final Node root = new Node(Integer.MIN_VALUE);
	volatile boolean stop = false;
	private MaintenanceThread mainThd;
	// used in the getSize function
	int size;

	// Constructors
	public OTBTreeOptimizedBackoff() {
		// temporary
		this.startMaintenance();
	}

	Node getNext(Node n, int k) {
		int rem = n.removed;
		if (rem == BY_LEFT_ROT)
			return n.right;
		else if (rem == REMOVED)
			return n.left;
		else if (n.key > k)
			return n.right;
		else if (n.key < k)
			return n.left;
		else
			// n.key == k
			return null;
	}

	int validate(Node n, int k, boolean itemExists) {
		Node next = null;
		if (n.removed != NOT_REMOVED)
			return STRUCTURALLY_INVALID;
//			return SEMANTICALLY_INVALID;
		else if (n.key == k) {
			if (n.deleted != itemExists)
				return VALID;
			else {
//				System.out.println("CANNOT HAPPEN IN THIS CASE\n");
				return SEMANTICALLY_INVALID;
			}
		} else if (n.key > k)
			next = n.right;
		else
			next = n.left;

		if (next == null) {
			if (!itemExists)
				return VALID;
			else {
//				System.out.println("CANNOT HAPPEN IN THIS CASE\n");
				return SEMANTICALLY_INVALID;
			}
		}
//		return SEMANTICALLY_INVALID;
		return STRUCTURALLY_INVALID;
	}

	class ReadSetEntry {
		Node node;
		int k;
		boolean itemExists;
		boolean inWriteSet;

		public ReadSetEntry(Node node, int k, boolean itemExists,
				boolean inWriteSet) {
			this.node = node;
			this.k = k;
			this.itemExists = itemExists;
			this.inWriteSet = inWriteSet;
		}
	}

	class WriteSetEntry {
		Node node;
		int k;
		Node newNode;
		int operation;

		public WriteSetEntry(Node node, int k, int operation) {
			this.node = node;
			this.k = k;
			this.operation = operation;
		}
	}

	@Override
	public boolean add(int item) throws AbortedException {
		return operation(ADD, item);
	}

	@Override
	public boolean remove(int item) throws AbortedException {
		return operation(REMOVE, item);
	}

	@Override
	public boolean contains(int item) throws AbortedException {
		return operation(CONTAINS, item);
	}

	private boolean commitValidate(ArrayList<ReadSetEntry> readset,
			TreeMap<Integer, WriteSetEntry> writeset, long threadId) {
		return postValidate(readset, writeset, threadId, true);
	}

	private boolean postValidate(ArrayList<ReadSetEntry> readset,
			TreeMap<Integer, WriteSetEntry> writeset, long threadId,
			boolean commitValidate) {
		ReadSetEntry readEntry;
		WriteSetEntry writeEntry;
		int size = readset.size();
		Node node, next;
		int lock;
		long lockHolder;

		// check the values of the nodes
		for (int i = 0; i < size; i++) {
			readEntry = readset.get(i);
			while (true) {

				lock = (readEntry.node.key == readEntry.k) ? readEntry.node.semantic_lock
						.get() : readEntry.node.struct_lock.get();
				lockHolder = (readEntry.node.key == readEntry.k) ? readEntry.node.semantic_lockHolder
						: readEntry.node.struct_lockHolder;

				if (lock == LOCKED && lockHolder != threadId) {
					if (lockHolder != MAINTAINANCE_THREAD_ID)
						return false;
					// else
					// while(lockHolder == MAINTAINANCE_THREAD_ID);
				}
				int validation_result = validate(readEntry.node, readEntry.k,
						readEntry.itemExists);
				if (validation_result == SEMANTICALLY_INVALID)
					return false;
				else if (validation_result == STRUCTURALLY_INVALID) {
					node = readEntry.node;
					while (true) {
						next = getNext(node, readEntry.k);
						if (next == null) {
							break;
						} else
							node = next;
					}
					readEntry.node = node;
					if (readEntry.inWriteSet) {
						writeEntry = writeset.get(readEntry.k);
						if (writeEntry != null)
							writeEntry.node = node;
					}

				} else
					break;
			}
		}

		return true;
	}

	private boolean operation(int type, int item) throws AbortedException {

		SetThread t = ((SetThread) Thread.currentThread());
		TreeMap<Integer, WriteSetEntry> writeset = t.tree_optimizedbackoff_writeset;
		ArrayList<ReadSetEntry> readset = t.tree_optimizedbackoff_readset;
		long threadId = Thread.currentThread().getId();

		WriteSetEntry entry = writeset.get(item);
		if (entry != null) {
			if (entry.operation == ADD) {
				if (type == ADD)
					return false;
				else if (type == CONTAINS)
					return true;
				else // remove
				{
					// System.out.println("remove after add");
					writeset.remove(item);
					return true;
				}
			} else // remove
			{
				if (type == REMOVE || type == CONTAINS)
					return false;
				else // add
				{
					// System.out.println("add after remove");
					writeset.remove(item);
					return true;
				}
			}
		}

		Node next, node;
		node = root;
		while (true) {
			next = getNext(node, item);
			if (next == null)
				break;
			node = next;
		}

		int currItem = node.key;
		boolean currDeleted = node.deleted;

		boolean returnValue = false;

		if (currItem == item && !currDeleted) {

			if (type == CONTAINS) {
				readset.add(new ReadSetEntry(node, item, true, false));
				returnValue = true;
			} else if (type == ADD) {
				readset.add(new ReadSetEntry(node, item, true, false));
				returnValue = false;
			} else // remove
			{
				readset.add(new ReadSetEntry(node, item, true, true));
				writeset.put(item, new WriteSetEntry(node, item, REMOVE));
				returnValue = true;
			}
		} else {
			if (type == CONTAINS || type == REMOVE) {
				readset.add(new ReadSetEntry(node, item, false, false));
				returnValue = false;
			} else // add
			{
				readset.add(new ReadSetEntry(node, item, false, true));
				writeset.put(item, new WriteSetEntry(node, item, ADD));
				returnValue = true;
			}
		}

		if (!postValidate(readset, writeset, threadId, false))
			throw AbortedException.abortedException;

		return returnValue;
	}

	@Override
	public void commit() throws AbortedException {
		SetThread t = ((SetThread) Thread.currentThread());

		Set<Entry<Integer, WriteSetEntry>> write_set = t.tree_optimizedbackoff_writeset
				.entrySet();
		ArrayList<ReadSetEntry> read_set = t.tree_optimizedbackoff_readset;

		// read-only transactions do nothing
		if (write_set.isEmpty()) {
			read_set.clear();
			return;
		}

		long threadId = Thread.currentThread().getId();
		Iterator<Entry<Integer, WriteSetEntry>> iterator = write_set.iterator();
		WriteSetEntry entry;

		AtomicInteger lock;
		long lockHolder;
		Node node, next, newNode;

		// Acquire locks
		while (iterator.hasNext()) {
			entry = iterator.next().getValue();

			while (true) {
				int lockType = (entry.node.key == entry.k) ? SEMANTIC_LOCK
						: STRUCTURAL_LOCK;
				lock = (entry.node.key == entry.k) ? entry.node.semantic_lock
						: entry.node.struct_lock;
				lockHolder = (entry.node.key == entry.k) ? entry.node.semantic_lockHolder
						: entry.node.struct_lockHolder;

				if (lockHolder == threadId)
					break;

				// Try to acquire lock

				if (lock.compareAndSet(UNLOCKED, LOCKED)) {
					if (lockType == SEMANTIC_LOCK)
						entry.node.semantic_lockHolder = threadId;
					else
						entry.node.struct_lockHolder = threadId;

					// validate the node. After this step, no need to validate
					// the entry again.
					boolean itemExists = false;
					if (entry.operation == REMOVE)
						itemExists = true;
					int validation_result = validate(entry.node, entry.k,
							itemExists);
					// valid. Done with this entry.
					if (validation_result == VALID)
						break;
					// structurally invalid. Unlock the node and get the new
					// valid node and re-validate
					else if (validation_result == STRUCTURALLY_INVALID) {
						if (lockType == SEMANTIC_LOCK)
							entry.node.semantic_lockHolder = threadId;
						else
							entry.node.struct_lockHolder = threadId;
						lock.set(UNLOCKED);

						node = entry.node;
						while (true) {
							next = getNext(node, entry.k);
							if (next == null) {
								break;
							} else
								node = next;
						}
						entry.node = node;
					} else
						// semantically invalid. Abort
						throw AbortedException.abortedException;

				} else if (lockHolder != MAINTAINANCE_THREAD_ID)
					throw AbortedException.abortedException;
			}
		}

		// validate read-set
		if (!commitValidate(t.tree_optimizedbackoff_readset,
				t.tree_optimizedbackoff_writeset, threadId))
			throw AbortedException.abortedException;

		// Publish write-set
		iterator = write_set.iterator();
		while (iterator.hasNext()) {
			entry = iterator.next().getValue();

			if (entry.operation == ADD) {
				if (entry.node.key == entry.k) {
					if (!entry.node.deleted
							|| entry.node.removed != NOT_REMOVED)
						System.out.println("CANNOT HAPPEN\n");
					entry.node.deleted = false;
				} else {
					newNode = new Node(entry.k);
					newNode.semantic_lock.set(LOCKED);
					newNode.semantic_lockHolder = threadId;
					newNode.struct_lock.set(LOCKED);
					newNode.struct_lockHolder = threadId;
					entry.newNode = newNode;

					// keep iterating. The only case here is that other
					// nodes
					// are inserted in the same transaction.
					node = entry.node;
					while (true) {
						next = getNext(node, entry.k);
						if (next == null) {
							break;
						} else
							node = next;
					}

					if (node.key > entry.k)
						node.right = newNode;
					else
						node.left = newNode;
				}
			} else // remove
			{
				entry.node.deleted = true;
			}
		}

		// unlock
		iterator = write_set.iterator();
		while (iterator.hasNext()) {
			entry = iterator.next().getValue();
			if (entry.node.semantic_lockHolder == threadId) {
				entry.node.semantic_lockHolder = -1;
				entry.node.semantic_lock.set(UNLOCKED);
			}
			if (entry.node.struct_lockHolder == threadId) {
				entry.node.struct_lockHolder = -1;
				entry.node.struct_lock.set(UNLOCKED);
			}
			// Unlock newNode
			if (entry.operation == ADD && entry.newNode != null) {
				if (entry.newNode.semantic_lockHolder == threadId
						&& entry.newNode.struct_lockHolder == threadId) {
					entry.newNode.semantic_lockHolder = -1;
					entry.newNode.semantic_lock.set(UNLOCKED);
					entry.newNode.struct_lockHolder = -1;
					entry.newNode.struct_lock.set(UNLOCKED);
				} else
					System.out.println("CANNOT HAPPEN\n");
			}
		}

		// clear read- and write- sets
		read_set.clear();
		write_set.clear();
	}

	@Override
	public void abort() {

		SetThread t = ((SetThread) Thread.currentThread());

		// internally you must not call this method. It's sufficient to throw
		// AbortException
		// This method will be called from calling transaction as a part of its
		// abort subroutine
		// This way: Abort is handled by transaction not by data structure.

		// Unlock
		Iterator<Entry<Integer, WriteSetEntry>> iterator = t.tree_optimizedbackoff_writeset
				.entrySet().iterator();
		WriteSetEntry entry;
		long threadId = t.getId();
		while (iterator.hasNext()) {
			entry = iterator.next().getValue();
			if (entry.node.semantic_lockHolder == threadId) {
				entry.node.semantic_lockHolder = -1;
				entry.node.semantic_lock.set(UNLOCKED);
			}
			if (entry.node.struct_lockHolder == threadId) {
				entry.node.struct_lockHolder = -1;
				entry.node.struct_lock.set(UNLOCKED);
			}
		}

		// clear read- and write- sets
		t.tree_optimizedbackoff_readset.clear();
		t.tree_optimizedbackoff_writeset.clear();

	}

	// maintenance

	boolean maintenanceLock(Node node, int type) {
		if (type == STRUCTURAL_LOCK) {
			if (node.struct_lock.compareAndSet(UNLOCKED, LOCKED)) {
				node.struct_lockHolder = MAINTAINANCE_THREAD_ID;
				return true;
			}
			return false;
		}

		if (type == SEMANTIC_LOCK) {
			if (node.semantic_lock.compareAndSet(UNLOCKED, LOCKED)) {
				node.semantic_lockHolder = MAINTAINANCE_THREAD_ID;
				return true;
			}
			return false;
		}

		// both locks
		if (node.semantic_lock.compareAndSet(UNLOCKED, LOCKED)) {
			node.semantic_lockHolder = MAINTAINANCE_THREAD_ID;
			if (node.struct_lock.compareAndSet(UNLOCKED, LOCKED)) {
				node.struct_lockHolder = MAINTAINANCE_THREAD_ID;
				return true;
			} else {
				node.semantic_lockHolder = -1;
				node.semantic_lock.set(UNLOCKED);
			}
		}
		return false;
	}

	void maintenanceUnlock(Node node) {
		if (node.semantic_lockHolder == MAINTAINANCE_THREAD_ID) {
			node.semantic_lockHolder = -1;
			node.semantic_lock.set(UNLOCKED);
		}
		if (node.struct_lockHolder == MAINTAINANCE_THREAD_ID) {
			node.struct_lockHolder = -1;
			node.struct_lock.set(UNLOCKED);
		}
	}

	boolean removeNode(Node parent, char direction) {
		Node n, child;
		// can get before locks because only maintenance removes nodes
		if (parent.removed != NOT_REMOVED)
			return false;
		n = direction == Left ? parent.left : parent.right;
		if (n == null)
			return false;

		// get the locks
		if (n.semantic_lock.get() == LOCKED || n.struct_lock.get() == LOCKED
				|| parent.struct_lock.get() == LOCKED)
			return false;
		
		if (maintenanceLock(parent, STRUCTURAL_LOCK)) {
			if (!maintenanceLock(n, BOTH_LOCKS)) {
				maintenanceUnlock(parent);
				return false;
			}
		} else
			return false;

		if (!n.deleted) {
			maintenanceUnlock(n);
			maintenanceUnlock(parent);
			return false;
		}
		if ((child = n.left) != null) {
			if (n.right != null) {
				maintenanceUnlock(n);
				maintenanceUnlock(parent);
				return false;
			}
		} else {
			child = n.right;
		}
		if (direction == Left) {
			parent.left = child;
		} else {
			parent.right = child;
		}
		n.left = parent;
		n.right = parent;

		num_removes++;

		n.removed = REMOVED;
		maintenanceUnlock(n);
		maintenanceUnlock(parent);

		// System.out.println("removed a node");

		// need to update balance values here
		if (direction == Left) {
			parent.bal.lefth = n.bal.localh - 1;
		} else {
			parent.bal.righth = n.bal.localh - 1;
		}
		parent.updateLocalh();
		return true;
	}

	int rightRotate(Node parent, char direction, boolean doRotate) {

		Node n, l, lr, r, newNode;
		if (parent.removed != NOT_REMOVED)
			return 0;
		n = direction == Left ? parent.left : parent.right;
		if (n == null)
			return 0;
		l = n.left;
		if (l == null)
			return 0;
		if (l.bal.lefth - l.bal.righth < 0 && !doRotate) {
			// should do a double rotate
			return 2;
		}

		if (n.semantic_lock.get() == LOCKED || n.struct_lock.get() == LOCKED
				|| l.struct_lock.get() == LOCKED)
			return 0;

		if (n.right == null || l.right == null || l.left == null) {
			// maintenanceLock(parent);
			if (maintenanceLock(n, BOTH_LOCKS)) {
				if (!maintenanceLock(l, STRUCTURAL_LOCK)) {
					maintenanceUnlock(n);
					return 0;
				}
			} else
				return 0;
		} else {
			if (!maintenanceLock(n, SEMANTIC_LOCK))
				return 0;
		}
		lr = l.right;
		r = n.right;
		newNode = new Node(n.key, Math.max(1 + l.bal.righth, 1 + n.bal.righth),
				l.bal.righth, n.bal.righth, lr, r, n.removed, n.deleted);
		l.right = newNode;

		// num_rotations++;

		n.removed = REMOVED;
		if (direction == Left) {
			parent.left = l;
		} else {
			parent.right = l;
		}
		maintenanceUnlock(l);
		maintenanceUnlock(n);
		// maintenanceUnlock(parent, false);
		// need to update balance values
		l.bal.righth = newNode.bal.localh;
		l.updateLocalh();
		if (direction == Left) {
			parent.bal.lefth = l.bal.localh;
		} else {
			parent.bal.righth = l.bal.localh;
		}
		parent.updateLocalh();
		// System.out.println("right rotate");
		return 1;

	}

	int leftRotate(Node parent, char direction, boolean doRotate) {
		Node n, r, rl, l, newNode;
		if (parent.removed != NOT_REMOVED)
			return 0;
		n = direction == Left ? parent.left : parent.right;
		if (n == null)
			return 0;
		r = n.right;
		if (r == null)
			return 0;
		if (r.bal.lefth - r.bal.righth > 0 && !doRotate) {
			// should do a double rotate
			return 3;
		}
		
		if (n.semantic_lock.get() == LOCKED || n.struct_lock.get() == LOCKED
				|| r.struct_lock.get() == LOCKED)
			return 0;

		if (n.left == null || r.right == null || r.left == null) {
			// maintenanceLock(parent);
			if (maintenanceLock(n, BOTH_LOCKS)) {
				if (!maintenanceLock(r, STRUCTURAL_LOCK)) {
					maintenanceUnlock(n);
					return 0;
				}
			} else
				return 0;
		} else {
			if (!maintenanceLock(n, SEMANTIC_LOCK))
				return 0;
		}

		rl = r.left;
		l = n.left;

		newNode = new Node(n.key, Math.max(1 + r.bal.lefth, 1 + n.bal.lefth),
				n.bal.lefth, r.bal.lefth, l, rl, n.removed, n.deleted);

		r.left = newNode;

		// // temp (Need to fix this!!!!!!!!!!!!!!!!!!!!)
		// n.right = parent;
		// n.left = parent;

		n.removed = BY_LEFT_ROT;

		num_rotations++;

		if (direction == Left) {
			parent.left = r;
		} else {
			parent.right = r;
		}
		maintenanceUnlock(r);
		maintenanceUnlock(n);
		// maintenanceUnlock(parent, false);
		// need to update balance values
		r.bal.righth = newNode.bal.localh;
		r.updateLocalh();
		if (direction == Left) {
			parent.bal.lefth = r.bal.localh;
		} else {
			parent.bal.righth = r.bal.localh;
		}
		parent.updateLocalh();
		// System.out.println("left rotate");
		return 1;
	}

	boolean propagate(Node node) {
		Node lchild, rchild;

		lchild = node.left;
		rchild = node.right;

		if (lchild == null) {
			node.bal.lefth = 0;
		} else {
			node.bal.lefth = lchild.bal.localh;
		}
		if (rchild == null) {
			node.bal.righth = 0;
		} else {
			node.bal.righth = rchild.bal.localh;
		}

		node.updateLocalh();

		if (Math.abs(node.bal.righth - node.bal.lefth) >= 2)
			return true;
		return false;
	}

	boolean performRotation(Node parent, char direction) {
		int ret;
		Node node;

		ret = singleRotation(parent, direction, false, false);
		if (ret == 2) {
			// Do a LRR
			node = direction == Left ? parent.left : parent.right;
			ret = singleRotation(node, Left, true, false);
			if (ret > 0) {
				if (singleRotation(parent, direction, false, true) > 0) {
					// System.out.println("LRR");
				}
			}
		} else if (ret == 3) {
			// Do a RLR
			node = direction == Left ? parent.left : parent.right;
			ret = singleRotation(node, Right, false, true);
			if (ret > 0) {
				if (singleRotation(parent, direction, true, false) > 0) {
					// System.out.println("RLR");
				}
			}
		}
		if (ret > 0)
			return true;
		return false;
	}

	int singleRotation(Node parent, char direction, boolean leftRotation,
			boolean rightRotation) {
		int bal, ret = 0;
		Node node, child;

		node = direction == Left ? parent.left : parent.right;
		bal = node.bal.lefth - node.bal.righth;
		if (bal >= 2 || rightRotation) {
			// check reiable and rotate
			child = node.left;
			if (child != null) {
				if (node.bal.lefth == child.bal.localh) {
					ret = rightRotate(parent, direction, rightRotation);
				}
			}
		} else if (bal <= -2 || leftRotation) {
			// check reliable and rotate
			child = node.right;
			if (child != null) {
				if (node.bal.righth == child.bal.localh) {
					ret = leftRotate(parent, direction, leftRotation);
				}
			}
		}
		return ret;
	}

	boolean recursivePropagate(Node parent, Node node, char direction) {

		Node left, right;

		if (node == null)
			return true;
		left = node.left;
		right = node.right;

		if (node.removed == NOT_REMOVED && node.deleted
				&& (left == null || right == null) && node != this.root) {
			if (removeNode(parent, direction)) {
				vars.rotations++;
				return true;
			}
		}

		if (stop) {
			return true;
		}

		if (node.removed == NOT_REMOVED) {
			if (left != null) {
				recursivePropagate(node, left, Left);
			}
			if (right != null) {
				recursivePropagate(node, right, Right);
			}
		}

		if (stop) {
			return true;
		}

		// no rotations for now
		if (node.removed == NOT_REMOVED && node != this.root) {
			if (propagate(node)) {
				this.performRotation(parent, direction);
				vars.rotations++;
			}
		}

		return true;
	}

	public boolean stopMaintenance() {
		this.stop = true;
		try {
			this.mainThd.join();
		} catch (InterruptedException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		System.out.println("\nNumber of rotations: " + num_rotations
				+ "\nNumber of physical removals: " + num_removes + "\n");
		return true;
	}

	public boolean startMaintenance() {
		this.stop = false;

		mainThd = new MaintenanceThread(this);

		mainThd.start();

		return true;
	}

	boolean doMaintenance() {
		long depth = 0, num_iterations = 0;
		long MAX_TIMEOUT = 1024, MIN_TIMEOUT = 2; //MICRO SECONDS
		long timeout = 1024;
		while (!stop) {
			long oldRotations = vars.rotations;
			vars.rotations= 0;
			recursivePropagate(this.root, this.root.left, Left);
			vars.propogations++;
			if(vars.rotations < oldRotations && timeout*2 < MAX_TIMEOUT)
					timeout*=2;
			else if(vars.rotations > oldRotations && timeout/2 > MIN_TIMEOUT)
					timeout/=2;
			try {
				TimeUnit.MICROSECONDS.sleep(timeout);
			} catch (InterruptedException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
			depth += recursiveDepth(this.root);
			num_iterations++;
		}
		System.out.println("NUMBER OF PROPAGATIONS IS: " + vars.propogations + "\n");
		System.out.println("NUmber of iterations: " + num_iterations + "average depth: " + depth/num_iterations + "\n");		
		return true;
	}

	// not thread safe
	public int getSize() {
		this.size = 0;
		recursiveGetSize(root.left);
		return size;
	}

	void recursiveGetSize(Node node) {
		if (node == null)
			return;
		if (node.removed != NOT_REMOVED) {
			// System.out.println("Shouldn't find removed nodes in the get size function");
		}
		if (!node.deleted) {
			this.size++;
		}
		recursiveGetSize(node.left);
		recursiveGetSize(node.right);
	}

	public int getBalance() {
		int lefth = 0, righth = 0;
		if (root.left == null)
			return 0;
		lefth = recursiveDepth(root.left.left);
		righth = recursiveDepth(root.left.right);
		return lefth - righth;
	}

	int recursiveDepth(Node node) {
		if (node == null) {
			return 0;
		}
		int lefth, righth;
		lefth = recursiveDepth(node.left);
		righth = recursiveDepth(node.right);
		return Math.max(lefth, righth) + 1;
	}

	public void clear() {
		this.stopMaintenance();
		this.resetTree();
		this.startMaintenance();

		return;
	}

	private void resetTree() {
		this.vars.propogations = 0;
		this.vars.rotations = 0;
		root.left = null;
	}

	@Override
	public void begin() {
		// TODO Auto-generated method stub

	}

	// This is a trick to insert the items balanced from the beginning
	// It's important to save the unnecessary time needed to re-balance
	// which is significant if the range is large.
	public boolean nontransactionalAdd(int key) {
		Node next, node;
		node = root;

		while (true) {
			next = getNext(node, key);
			if (next == null) {
				if (node.struct_lock.compareAndSet(UNLOCKED, LOCKED)) {
					if (validate(node, key, false) == VALID)
						break;
					else
						node.struct_lock.set(UNLOCKED);
				}
			} else
				node = next;
		}

		boolean result = false;

		if (node.key == key) {
			if (node.deleted) {
				node.deleted = false;
				result = true;
			}
		} else {
			Node n = new Node(key);
			if (node.key > key)
				node.right = n;
			else
				node.left = n;
			result = true;
		}
		node.struct_lock.set(UNLOCKED);
		return result;
	}

	public void recprint(Node p) {
		if (p == null)
			return;
		recprint(p.right);
		System.out.print(p.key + "  ");
		recprint(p.left);
	}

	public int recgetvalue(Node p) {
		if (p == null)
			return 0;
		if (p.deleted || p == root)
			return recgetvalue(p.right) + recgetvalue(p.left);
		else
			return recgetvalue(p.right) + p.key + recgetvalue(p.left);
	}

	public void recursiveAdd(int start, int end) {
		if (start > end)
			return;

		int mid = (start + end) / 2;
		nontransactionalAdd(mid * 2);
		recursiveAdd(start, mid - 1);
		recursiveAdd(mid + 1, end);

		return;
	}

	@Override
	public int finish() {
		// TODO Auto-generated method stub
		stopMaintenance();
		return recgetvalue(root);
	}

	@Override
	public void start(int range) {
		// TODO Auto-generated method stub
		recursiveAdd(0, range / 2 - 1);

		// recprint(root);
		// System.out.println();
	}

}
